unit stega;

interface

uses Windows, KOL,binary;
type

(*      

    TDataRecord = record
       .............
       dLabel : string[5];             ()
       dCountBit : byte;              (1..8)
       dCompress : byte;
       dCrypt    : byte;
       dCountFiles : integer;        
       ....... 1........      ------------------------------
       dFileSize : longint;         
       dLengthFilename: byte;        
       dFileName : string;          
       dFileData : array of byte;   
       ....... 2........      -----------------------------
    end;                     *)

    THeaderStg = record
      case byte of
      0 : (
            dLabel : string[5];
            dCountBit : byte;
            dCompress : byte;
            dCrypt    : byte;
            dCountFiles : integer; );
      1 : ( ddd : array[0..15] of byte);
    end;

    //      
    TFileInfo = packed record
       iFilename : string;   // 
       iFileSize : longint;  // 
       iDataPos  : cardinal; //     
    end;

    //  
    PStegaBmp =^TStegaBmp;
    TStegaBmp = object( TObj )
    protected
       destructor Destroy; virtual;
    private
       FCountDataBit   : byte;     //    (1...7)
       FCRCError       : boolean;  // : / CRC 
       FDataExist      : boolean;  // : / 
       FImageFileName  : string;   //   
       FLabelStega     : string;   // 
       FContainerSize  : longint;  //  
       FImageStream    : pStream;
       FDataStream     : pStream;
       FListFiles      : pList;    //  
       //---------------------------------------
       function GetSizeContainer : integer;
       function GetFreeSizeContainer : integer;
       function GetCountFiles : integer;
       procedure SetCountDataBit(const val:byte);
       procedure SetLabelStega(const val:string);
       //---------------------------------------
       function ReadByteFromImage : byte;
       function ReadIntegerFromImage : integer;
       function ReadStringFromImage(len:byte):string;
       procedure WriteByteToImage(val:byte);
       function CalcDataCRC32 : DWord;
       //---------------------------------------
       function CheckRightImage : boolean;
       procedure SeekBmp0;
       //---------------------------------------
    public
       CompressData : boolean;     //  
       CryptData    : boolean;     //  
       function GetFileName(idx:integer):string;
       function GetFileSize(idx:integer):longint;
       function GetFileIndex(FileName:string):integer;
       //---------------------------------------
       function  LoadImage(const FileName:string):boolean;
       procedure SaveImage(const FileName:string);
       function  AddFile(const FileName:string):boolean;
       procedure ExtractFile(idx:integer;const OutFileName:string);
       //---------------------------------------
       property LabelStega : string read FLabelStega write SetLabelStega;
       property CountDataBit : byte read FCountDataBit write SetCountDataBit;
       property CRCError : boolean read FCRCError;
       property DataExist : boolean read FDataExist;
       property FileImage : string read FImageFileName;
       property CountFiles : integer read GetCountFiles;
       property SizeContainer : integer read GetSizeContainer;
       property FreeSizeContainer : integer read GetFreeSizeContainer;
    end;

    function NewStegaBmp: PStegaBmp;

implementation

{ TStegoBMP }

function NewStegaBmp;
begin
     New(Result, Create);
     Result.FImageStream:=NewMemoryStream;
     Result.FDataStream:=NewMemoryStream;
     Result.FListFiles := NewList;
     Result.FCountDataBit:=1;
     Result.FLabelStega:='STEGA';
     Result.FCRCError:=false;
     Result.FDataExist:=false;
     Result.CompressData:=false;
     Result.CryptData:=false;
end;

destructor TStegaBMP.Destroy;
begin
     FListFiles.Release;
     FDataStream.Free;
     FImageStream.Free;
     inherited;
end;

function TStegaBmp.GetSizeContainer: integer;
begin
      Result:=((FContainerSize*FCountDataBit) div 8) //  
               -SizeOf(THeaderStg);                  // - 
end;

function TStegaBmp.GetFreeSizeContainer: integer;
var
   i  : integer;
   sz : longint;
   t  : ^TFileInfo;
begin
     sz:=0;
     for i:=0 to FListFiles.Count-1 do
     begin
       t:=FListFiles.Items[i];
       //  [Fs][Fn] Fn                   Fs
       sz:=sz+4+1+Length(t^.iFilename)+t^.iFileSize;
     end;
     Result:=SizeContainer-sz;
     If Result<0 then Result:=0;
end;

function TStegaBmp.GetCountFiles: integer;
begin
     Result:=FListFiles.Count;
end;

procedure TStegaBmp.SetCountDataBit(const val: byte);
begin
     If (val>0) and (val<9) then FCountDataBit:=val;
end;

procedure TStegaBmp.SetLabelStega(const val:string);
begin
     if val<>'' then FLabelStega:=val;
end;

function TStegaBmp.ReadByteFromImage: byte;
var
   i,n     : integer;
   Bits    : TBitsArray;
   src,tmp : byte;
begin

  n:=FCountDataBit-1;

  for i:=7 downto 0 do
  begin

    if n=FCountDataBit-1 then FImageStream.Read(src,1);

    tmp:=src;

    if (tmp and (1 shl n))<>0
      then tmp:=1
      else tmp:=0;

    Bits[i]:=tmp;

    if n=0
       then n:=FCountDataBit-1
       else n:=n-1;

  end;
  Result:=BinToHex(Bits);
end;

function TStegaBmp.ReadIntegerFromImage: integer;
var
    b : TDoubleWordArray;
    i : integer;
begin
   for i:=0 to 3 do b[i]:=ReadByteFromImage;
   Result:=z4BytesToInt(b);
end;

function TStegaBmp.ReadStringFromImage(len: byte): string;
var
   i : integer;
begin
   Result:='';
   for i:=1 to len do Result:=Result+Chr(ReadByteFromImage);
end;

procedure TStegaBmp.WriteByteToImage(val: byte);
var
   i,n   : integer;
   Bits  : TBitsArray;
   src   : byte;
begin
     n:=FCountDataBit-1;

     HexToBin(val,Bits);

     for i:=7 downto 0 do
     begin

       if N=FCountDataBit-1 then FImageStream.Read(src,1);

       if Bits[i]=1
         then src:=SetBit(src,N)
         else src:=ResetBit(src,N);

       if N=0 then N:=FCountDataBit-1
              else N:=N-1;

       FImageStream.Position:=FImageStream.Position-1;
       FImageStream.Write(src,1);

     end;
end;

function TStegaBmp.CalcDataCRC32: DWord;

{$I stega.inc}

var
   i        :  integer;
   CRCval   :  DWord;
   b        : byte;
begin
     FDataStream.Position:=0;
     CRCVal := $FFFFFFFF;
     for i:=0 to FDataStream.Size-1 do
     begin
       FDataStream.Read(b,1);
       CRCval := (CRCval SHR 8) XOR Table[ b XOR (CRCval AND $000000FF) ];
     end;
     Result := NOT CRCval;
end;

procedure TStegaBmp.SeekBmp0;
var
    bfh : TBitmapFileHeader;
begin
     FImageStream.Position:=0;
     FImageStream.Read(BFH,Sizeof(BFH));
     FImageStream.Position:=bfh.bfOffBits;
end;

function TStegaBmp.CheckRightImage : boolean;
var
    wrd  : word;
    dwrd : dword;
begin
     Result:=true;
     FImageStream.Position:=0;
     FImageStream.Read(wrd,2);
     if wrd<>$4D42 then Result:=false;   // 
     FImageStream.Position:=FImageStream.Position+26;
     FImageStream.Read(wrd,2);
//     If wrd<>$18 then Result:=false;     // 24 
     FImageStream.Read(dwrd,4);
     if dwrd<>0 then Result:=false;      // 
end;

function TStegaBmp.GetFileName(idx: integer): string;
var
    t : ^TFileInfo;
begin
     t:=FListFiles.Items[idx];
     Result:=t^.iFilename;
end;

function TStegaBmp.GetFileSize(idx: integer): longint;
var
    t : ^TFileInfo;
begin
     t:=FListFiles.Items[idx];
     Result:=t^.iFileSize;
end;

function TStegaBmp.GetFileIndex(FileName: string): integer;
var
    t : ^TFileInfo;
    i : integer;
begin
     Result:=-1;
     for i:=0 to FListFiles.Count-1 do
     begin
       t:=FListFiles.Items[i];
       if t^.iFilename=FileName then
       begin
         Result:=i;
         Exit;
       end;
     end;
end;

function TStegaBmp.LoadImage(const FileName: string):boolean;
var
   v         : byte;
   i,j,n     : integer;
   crc0,crc1 : DWord;
   tmpStream : pStream;
   t         : ^TFileInfo;
   h         : THeaderStg;
   fnm       : string;
   function ReadHeaderStegas(var hdr:THeaderStg):boolean;
   var
      i : integer;
   begin
     Result:=false;
     for i:=0 to SizeOf(hdr)-1 do hdr.ddd[i]:=ReadByteFromImage;
     if hdr.dLabel=FLabelStega then Result:=true;
   end;
begin
     //   
     Result:=false;
     FDataExist:=false;
     FCRCError:=false;
     FDataStream.Size:=0;
     FImageStream.Size:=0;
     FListFiles.Clear;
     FImageFileName:=FileName;
     //    
     tmpStream := NewReadFileStream(FileName);
     Stream2Stream(FImageStream,tmpStream,tmpStream.Size);
     tmpStream.Free;
     //        
      If Not CheckRightImage then Exit;

     //    
     SeekBmp0;
     //     
     FContainerSize:=FImageStream.Size-FImageStream.Position;

     //  
     v:=FCountDataBit;
     //    1 
     FCountDataBit:=1;
     if ReadHeaderStegas(h) then
     begin

       FDataStream.Write(h,SizeOf(h));

       FDataExist:=true;               //      
       FCountDataBit:=h.dCountBit;

       If h.dCountFiles>255 then
       begin
         MsgBox('Data corrupt',mb_OK+mb_IconError);
         Exit
       end;

       for i:=0 to h.dCountFiles-1 do
       begin

         New(t);

         N:=ReadIntegerFromImage;
         FDataStream.Write(N,4);

         v:=ReadByteFromImage;
         FDataStream.Write(v,1);

         fnm:=ReadStringFromImage(v);
         FDataStream.WriteStr(fnm);

         t^.iFilename:=fnm;
         t^.iFileSize:=N;
         t^.iDataPos:=FDataStream.Position;
         FListFiles.Add(t);

         for J:=0 to N-1 do
         begin
           v:=ReadByteFromImage;
           FDataStream.Write(v,1);
         end;
       end;

       crc0:=ReadIntegerFromImage;
       crc1:=CalcDataCRC32;

       if crc0<>crc1 then FCRCError:=true;
     end
     else FCountDataBit:=v;
     Result:=true;
end;

procedure TStegaBmp.SaveImage(const FileName: string);
var
   v         : byte;
   fnm       : string;
   i,N       : integer;
   crc32     : dword;
   t         : ^TFileInfo;
   tmpStream : pStream;
   hdr       : THeaderStg;
begin
     FImageFileName:=FileName;
     FDataStream.Position:=0;

     //-----------   ------------------
    hdr.dLabel:=FLabelStega;
    hdr.dCountBit:=FCountDataBit;
    if CompressData
       then hdr.dCompress:=1
       else hdr.dCompress:=0;
    If CryptData
       then hdr.dCrypt:=1
       else hdr.dCrypt:=0;
    hdr.dCountFiles:=FListFiles.Count;
    FDataStream.Write(hdr,SizeOf(hdr));

    //         
    IF FDataExist then FDataStream.Position:=FDataStream.Size;

    //   
    for i:=0 to FListFiles.Count-1 do
    begin
{*}   t:=FListFiles.Items[i];

      if t^.iDataPos>0 then Continue;         //    ...

      tmpStream:=NewReadFileStream(t^.iFilename);
      fNm:=ExtractFileName(t^.iFilename);

      //--------------   ------------------------
      N:=tmpStream.size;
      FDataStream.Write(N,4);

      v:=Length(fNm);
      FDataStream.Write(v,1);

      FDataStream.WriteStr(fnm);

{*}   t^.iDataPos:=FDataStream.Position;

      //   
      Stream2Stream(FDataStream,tmpStream,tmpStream.Size);

      //   
      tmpStream.Free;
    end;
    if CompressData then
    begin
         ;
    end;
    if CryptData then
    begin
         ;
    end;
    //---      -----------------------------
    crc32:=CalcDataCRC32;
    FDataStream.Write(crc32,4);
    //--------    -----------------------------------
    SeekBMP0;
    FDataStream.Position:=0;

    //     1 
    FCountDataBit:=1;
    for i:=0 to SizeOf(hdr)-1 do
    begin
      FDataStream.Read(v,1);
      WriteByteToImage(v);
    end;
    //  ;
    FCountDataBit:=hdr.dCountBit;
    for i:=SizeOf(hdr) to FDataStream.Size do
    begin
      FDataStream.Read(v,1);
      WriteByteToImage(v);
    end;
    //-------------------     ---------------------------
    FImageStream.Position:=0;
    tmpStream:=NewWriteFileStream(FileName);
    Stream2Stream(tmpStream,FImageStream,FImageStream.Size);
    tmpStream.Free;
    FDataExist:=true;
end;

function TStegaBmp.AddFile(const FileName: string):boolean;
var
   fsz : longint;
   t : ^TFileInfo;
begin
     Result:=false;
     fsz:=FileSize(Filename);
     If fsz<=FreeSizeContainer then
     begin
       New(t);
       t^.iFilename:=FileName;
       t^.iFileSize:=fsz;
       t^.iDataPos:=0;
       FListFiles.Add(t);
       Result:=true;
     end;
end;

procedure TStegaBmp.ExtractFile(idx: integer; const OutFileName: string);
var
   i : integer;
   f : file;
   b : byte;
   t : ^TFileInfo;
begin
     t:=FListFiles.Items[idx];
     If t^.iDataPos=0 then exit;
     FDataStream.Position:=t^.iDataPos;
     AssignFile(f,OutFileName);
     Rewrite(f,1);
     for i:=0 to t^.iFileSize-1 do
     begin
       FDataStream.Read(b,1);
       BlockWrite(f,b,1);
     end;
     CloseFile(f);
end;

end.
